// Convert ProMC to HEPMC files 
// This code is experimental. It reads ProMC and tries to write HEPMC file.
// Author: Sergei Chekanov (ANL) 

#include "HepMC/GenEvent.h"
#include "HepMC/IO_GenEvent.h"
#include <cstdlib>
#include <string>
#include <cassert>
#include <iostream>
#include "ProMC.pb.h"
#include "ProMCBook.h"
#include "ProMCHeader.pb.h"

using namespace std;
using namespace HepMC;



// main program to read ProMC and convert to HEPMC
int main(int argc, char** argv) {

  // Look for a help argument
  for (int i = 1; i < argc; ++i) {
    const string arg = argv[i];
    if (arg == "--help" || arg == "-h") {
      cout << argv[0] << ": a ProMC to HepMC converter" << endl;
      cout << "Usage: " << argv[0] << "[<input ProMC file> <output HEPMC file>]" << endl;
      cout << "This program converts a ProMC file to HepMC" << endl;
      exit(0);
    }
  }

  // Choose input and output files from the command line
  string infile("-"), outfile("-");
  if (argc == 3) {
    infile = argv[1];
    outfile = argv[2];
  } else if (argc != 3) {
    cerr << "Usage: " << argv[0] << "[<PROMC files input> <HEPMC file out>]" << endl;
    exit(1);
  }

 // -------------- ProMC file ----------------------
  ProMCBook *epbook = new ProMCBook( infile.c_str(),"r");
  // get the version number
  int  h=epbook->getVersion();
  // cout << "ProMC version = " << h << endl;
  // get the description of this file
  //string d=epbook->getDescription();
  //cout << "Description = " << d << endl;
  unsigned int  nev=epbook->getEvents();
  cout << "Events = " << nev  << endl;

  // get the header file with units, cross sections etc.
  ProMCHeader header = epbook->getHeader();
  //cout << "Cross section=" << header.cross_section() << endl;
  // here is the units for energy
  float kEV=(float)(header.momentumunit());
  float kLe=(float)(header.lengthunit()); 
  cout << "Energy unit=" << kEV << endl;
  cout << "Length unit=" << kLe << endl;
 

 
  IO_GenEvent* writer = (outfile == "-") ? new IO_GenEvent(cout) : new IO_GenEvent(outfile, ios::out);
  writer->write_comment(epbook->getDescription());
 
 // HepMC::GenCrossSection xsec;
 // xsec.set_cross_section( header.cross_section());
 // evt->set_cross_section(xsec);



  for (unsigned int j=0; j<nev; j++){ // loop over events
  if (epbook->next() !=0) continue;
  ProMCEvent eve = epbook->get();


  // event information
  ProMCEvent_Event  *e=eve.mutable_event();
  int event_number=e->number();
  int signal_process_id=e->process_id();
  float event_scale=e->scale();
  float alphaQED =e->alpha_qed();
  float alphaQCD =e->alpha_qcd();
  int  id1pdf=e->id1();
  int  id2pdf=e->id2();
  float  x1pdf=e->x1();
  float  x2pdf=e->x2();
  float  pdf1=e->pdf1();
  float  pdf2=e->pdf2();
  float  Qfac=e->scale_pdf();

  // Construct new empty HepMC event.
  GenEvent* hepmcevt = new GenEvent( signal_process_id, event_number );
  // define the units
  hepmcevt->use_units(HepMC::Units::GEV, HepMC::Units::MM);
  hepmcevt->set_event_number(event_number);
  hepmcevt->set_signal_process_id(signal_process_id);
  hepmcevt->set_event_scale(event_scale);
  hepmcevt->set_alphaQED( alphaQED );
  hepmcevt->set_alphaQCD( alphaQCD );
  hepmcevt->set_pdf_info( PdfInfo( id1pdf, id2pdf, x1pdf, x2pdf, Qfac, pdf1, pdf2) );



  // get truth particle information 
  ProMCEvent_Particles  *pa=eve.mutable_particles();

  std::vector<GenParticle*> hepevt_particles( pa->pdg_id_size() );

  for (unsigned int n=0; n<pa->pdg_id_size(); n++){ // loop over particles
          int id=pa->id(n); // position in the record 
          int pid=pa->pdg_id(n); // PDG id 
          int status=pa->status(n);
          float px=(pa->px(n)/kEV); 
          float py=(pa->py(n)/kEV);
          float pz=(pa->pz(n)/kEV);
          float mass=(pa->mass(n)/kEV); 
          int daughter1=pa->daughter1(n);
          int daughter2=pa->daughter2(n);
          int mother1=pa->mother1(n);
          int mother2=pa->mother2(n);
          double e=sqrt(px*px+py*py+pz*pz+mass*mass);
         // fill HEPMC
          hepevt_particles[n] = new GenParticle(HepMC::FourVector(px,py,pz,e),pid,status); 
          hepevt_particles[n]->suggest_barcode(id);
          hepevt_particles[n]->set_generated_mass(mass); 
  }        

  hepmcevt->set_beam_particles( hepevt_particles[1], hepevt_particles[2] );

  // extract production vertex and fill hepmc 
  for (unsigned int n=0; n<pa->t_size(); n++){
    float x=(pa->x(n)/kLe);
    float y=(pa->y(n)/kLe);
    float z=(pa->z(n)/kLe);
    float t=(pa->t(n)/kLe);

    GenParticle *p = hepevt_particles[n];
    HepMC::FourVector prod_pos( x,y,z,t);
    GenVertex* prod_vtx = new GenVertex();
    prod_vtx->set_position( prod_pos );
    if ( prod_vtx ) prod_vtx->add_particle_out( p );
    hepmcevt->add_vertex( prod_vtx );

  }
  

  if (j%100==0)
          cout << j <<  " Nr particles=" << pa->pdg_id_size()  << endl;


  writer->write_event(hepmcevt); // write HEPMC event 


  } // end loop over events



 
  epbook->close(); // close
  delete writer;

  return EXIT_SUCCESS;
}
