// main02.cc is a part of the PYTHIA event generator.
// Copyright (C) 2009 Torbjorn Sjostrand.
// PYTHIA is licenced under the GNU GPL version 2, see COPYING for details.
// Please respect the MCnet Guidelines, see GUIDELINES for details.
// 
//  S.Chekanov (ANL). 
//  Add an example of how to write ProMC files   
#include <TROOT.h>
#include <TApplication.h>
#include <TTree.h>
#include <TFile.h>
#include <TH1D.h>
#include <TH2D.h>
#include <map>
#include "fastjet/PseudoJet.hh"
#include "fastjet/ClusterSequence.hh"
#include "Pythia.h"
#include "HepMCInterface.h"
#include "HepMC/GenEvent.h"
#include "HepMC/IO_GenEvent.h"

// Following line to be used with HepMC 2.04 onwards.
#ifdef HEPMC_HAS_UNITS
#include "HepMC/Units.h"
#endif

// ProMC file
#include "ProMCHeader.pb.h"
#include "ProMC.pb.h"
#include "ProMCBook.h"


using namespace Pythia8; 
using namespace fastjet;



int main() {

   

  // Generator. Process selection. Tevatron initialization. Histogram.
  Pythia pythia;


 // total events
 int Ntot=1000;


 // fill HEPMC for comparisons
 HepMC::I_Pythia8 ToHepMC;
 HepMC::IO_GenEvent ascii_io("Pythia8.hepmc", std::ios::out);

 // set some QCD events
  pythia.readString("Beams:eCM = 14000.");
  pythia.readString("HardQCD:all = on");
  pythia.readString("WeakSingleBoson:all");
  pythia.readString("WeakDoubleBoson:all = on");
  pythia.readString("Top:gg2ttbar");
  pythia.readString("Top:qqbar2ttbar=on");
  pythia.readString("Random:setSeed = on");
  pythia.readString("Random:seed = 0");
  pythia.readString("PhaseSpace:pTHatMin = 20.");
  pythia.init( 2212, 2212, 14000.);
  
  pythia.settings.listChanged(); // Show changed settings
  pythia.particleData.listChanged(); // Show changed particle data



   // ****************  book ProMC file **********************
  ProMCBook*  epbook = new ProMCBook("Pythia8.promc","w");
  epbook->setDescription(Ntot,"PYTHIA8 event record");
  // **************** Set a header ***************************
  ProMCHeader header;
  // cross section in pb 
  header.set_cross_section( pythia.info.sigmaGen() * 1e9 );
  header.set_cross_section_error( pythia.info.sigmaErr() * 1e9 );
  // the rest
  header.set_id1( pythia.info.idA() );
  header.set_id2( pythia.info.idB() );
  header.set_pdf1( pythia.info.pdf1() );
  header.set_pdf2( pythia.info.pdf2() );
  header.set_x1(  pythia.info.x1() );
  header.set_x2(  pythia.info.x2()  );
  header.set_scalepdf(  pythia.info.QFac()  );
  header.set_weight( pythia.info.weight());
  header.set_name(pythia.info.name());
  header.set_code(pythia.info.code());
  header.set_ecm(pythia.info.eCM());
  header.set_s(pythia.info.s());

 // Use the range 0.01 MeV to 20 TeV using varints (integers)  
 // if particle in GeV, we mutiple it by kEV, to get 0.01 MeV =1 unit
 const double kEV=1000*100;
 // if lenght is in CM, use 0.1 mm = 1 unit
 const double kL=1000;
  // set units in 0.01*MeV for momenta 
  header.set_momentumunit((int)kEV);
  header.set_lengthunit((int)kL);

  cout << "****  Store particle data in the header file.." << endl;
  for (int i=0; i<10000; i++) {
     if (pythia.particleData.name(i) != " ") {
     ProMCHeader_ParticleData* pp= header.add_particledata();
     pp->set_id(i); // pi+ 
     pp->set_mass(pythia.particleData.m0(i));
     pp->set_name(pythia.particleData.name(i));
     pp->set_width(pythia.particleData.mWidth(i));
     pp->set_lifetime(pythia.particleData.tau0(i));
     pp->set_charge(pythia.particleData.chargeType(i));
     cout << "Save mass=" << pythia.particleData.m0(i) << " " << pythia.particleData.name(i) << " for PID=" << i << endl;
     }
  }

  epbook->setHeader(header); // write header 

 
    // book a histogram, make sure that the output name is Analysis.root 
    char*  ffile="Pythia8.root";
    cout << "\n -> Output file is =" << ffile << endl;
    TFile * RootFile = new TFile(ffile, "RECREATE", "Histogram file");
    TH1D * h_pt = new TH1D("ptjet", "ptjet", 200, 250, 1000);


  // Begin event loop. Generate event. Skip if error. List first one.
  for (int n = 0; n < Ntot; n++) {
    if (!pythia.next()) continue;
    if (n < 3) {pythia.info.list(); pythia.event.list();}
    // Loop over particles in event. Find last Z0 copy. Fill its pT. 

     if (n<=10 &&
        n<=100 && (n%10) == 0 ||
        n<=1000 && (n%100) == 0  ||
        n>=1000 && (n%1000) == 0. ) {
        cout << "No events= " << n << " passed"  << endl; };

   // If failure because reached end of file then exit event loop.
      if (pythia.info.atEndOfFile()) {
        cout << " Aborted since reached end of Les Houches Event File\n";
        break;
      }


     // MCHep compiled it with GEV and MM
     HepMC::GenEvent* hepmcevt = new HepMC::GenEvent(
     HepMC::Units::GEV, HepMC::Units::MM);

/*
#ifdef HEPMC_HAS_UNITS
    // This form with arguments is only meaningful for HepMC 2.04 onwards, 
    // and even then unnecessary if HepMC was built with GeV and mm as units.
    HepMC::GenEvent* hepmcevt = new HepMC::GenEvent(
      HepMC::Units::GEV, HepMC::Units::MM);
#else
    // This form is needed for backwards compatibility. 
    // In HepMCInterface.cc a conversion from GeV to MeV will be done.
    HepMC::GenEvent* hepmcevt = new HepMC::GenEvent();
#endif
*/

    // Fill HepMC event, including PDF info.
    ToHepMC.fill_next_event( pythia, hepmcevt );
    ascii_io << hepmcevt;
    delete hepmcevt;



    

  //************  ProMC file ***************//
  ProMCEvent promc;

  // fill event information
  ProMCEvent_Event  *eve= promc.mutable_event();
  eve->set_number(n);
  eve->set_process_id(pythia.info.code());     // process ID 
  eve->set_scale(pythia.info.pTHat());       
  eve->set_alpha_qed(pythia.info.alphaEM()); 
  eve->set_alpha_qcd(pythia.info.alphaS()); 
  eve->set_scale_pdf(pythia.info.QFac());  
  eve->set_weight(pythia.info.weight());  
  eve->set_x1(pythia.info.x1pdf());  
  eve->set_x2(pythia.info.x2pdf()); 
  eve->set_id1(pythia.info.id1pdf()); 
  eve->set_id2(pythia.info.id2pdf()); 
 

  // fill truth particle information 
  ProMCEvent_Particles  *pa= promc.mutable_particles();

  for (int i =0; i<pythia.event.size(); i++) {
  pa->add_pdg_id( pythia.event[i].id() );
  pa->add_status(  pythia.event[i].status() );
  pa->add_px( (int)(pythia.event[i].px()*kEV) );
  pa->add_py(  (int)(pythia.event[i].py()*kEV) );
  pa->add_pz(   (int)(pythia.event[i].pz()*kEV)  );
  pa->add_mass( (int)(pythia.event[i].m()*kEV) ); // we also keep masses in header 
  pa->add_mother1( pythia.event[i].mother1()  );
  pa->add_mother2( pythia.event[i].mother2()  );
  pa->add_daughter1( pythia.event[i].daughter1()  );
  pa->add_daughter2( pythia.event[i].daughter2()   );
  pa->add_barcode( i  ); // dummy 
  pa->add_id( i  );
  pa->add_x( pythia.event[i].xProd()*kL  );
  pa->add_y( pythia.event[i].yProd()*kL   );
  pa->add_z( pythia.event[i].zProd()*kL  );
  pa->add_t( pythia.event[i].tProd()*kL   );


 // check conservartion
  double ipx= ((int)(pythia.event[i].px()*kEV))/(double)kEV;
  double ipy= ((int)(pythia.event[i].py()*kEV))/(double)kEV;
  double ipz= ((int)(pythia.event[i].pz()*kEV))/(double)kEV;
  double mm=  ((int)(pythia.event[i].m()*kEV))/(double)kEV;

/*
  if ( pythia.event[i].isFinal())   {
    //cout << pythia.event[i].id() << endl; 
     double ee1=ipx*ipx+ipy*ipy+ipz*ipz - mm*mm; 
     double ee2=pythia.event[i].px()*pythia.event[i].px()+pythia.event[i].py()*pythia.event[i].py()+pythia.event[i].pz()*pythia.event[i].pz()-pythia.event[i].m()*pythia.event[i].m(); 
     if (ee2<0) cout << ee2 << " " << " Negative! (float)" << " " << pythia.event[i].id() << endl; 
     if (ee1<0) cout << ee1 << " " << " Negative! (int)" << " " << pythia.event[i].id() << endl;

   }
*/

  // cout << pythia.event[i].xProd() << endl;
  //  cout << i << " kEv Input=" << (int)(pythia.event[i].pz()*kEV) << " Out=" << promc.pz(i) <<  endl;  
    }

    epbook->write(promc); // write event



  } // endl loop over events 


   // To check which changes have actually taken effect 
   pythia.settings.listChanged();
   // pythia.particleData.listChanged();
   pythia.particleData.list(25);
   // ParticleDataTable::listAll()
   // ParticleDataTable::list(25);


  pythia.statistics();

  // Output histograms
  double sigmapb = pythia.info.sigmaGen() * 1.0E9;
  cout << "== Cross section for this run=" <<  sigmapb << " pb" << endl; 
  cout << "== Events for this run=" <<  Ntot << endl;
  double lumi=(Ntot/sigmapb)/1000;
  cout << "== Luminosity for this run=" <<  lumi  << " fb-1" << endl;
  cout << "\n\n-- Write output file=" << ffile << endl;
  cout << "\n\n";

   RootFile->Write();
   RootFile->Print();
   RootFile->Close();

  
   cout << pythia.info.idA() << endl;


  // save post-generation statistics for ProMC
  ProMCStat stat;
  stat.set_cross_section_accumulated( sigmapb ); // in pb 
  stat.set_cross_section_error_accumulated( pythia.info.sigmaErr() * 1e9 );
  stat.set_luminosity_accumulated(  Ntot/sigmapb );
  stat.set_ntried(pythia.info.nTried());
  stat.set_nselected(pythia.info.nSelected());
  stat.set_naccepted(pythia.info.nAccepted());
  epbook->setStatistics(stat); 

  // close the ProMC file
  epbook->close(); // close



  return 0;
}

